@@ -1,4 +1,4 @@
-import os, re, subprocess
+import sys, os, re, subprocess
 
 from django.core.management.base import BaseCommand, CommandError
 
@@ -84,8 +84,30 @@ class Command(BaseCommand):
         self.project_root = settings.BASE_DIR
         self.settings_path = f"{self.project_root}/{self.project_name}/settings.py"
 
-        # Are we using requirements.txt?
+        # Which dependency management approach are we using?
+        #   req_txt, pipenv, poetry?
+        #   In a simple project, I don't think we should find both Pipfile
+        #   and requirements.txt. However, if there's both, prioritize Pipfile.
         self.using_req_txt = 'requirements.txt' in os.listdir(self.project_root)
+
+        # DEV: If both req_txt and Pipfile are found, could just use req.txt.
+        #      That's Heroku's prioritization, I believe. Address this if
+        #      anyone has such a project, and ask why they have both initially.
+        self.using_pipenv = 'Pipfile' in os.listdir(self.project_root)
+        if self.using_pipenv:
+            self.using_req_txt = False
+
+        self.using_poetry = 'pyproject.toml' in os.listdir(self.project_root)
+        if self.using_poetry:
+            # Heroku does not recognize pyproject.toml, so we'll export to
+            #   a requirements.txt file, and then work from that. This should
+            #   not affect the user's local environment.
+            export_cmd_parts = ['poetry', 'export', '-f', 'requirements.txt',
+                    '--output', 'requirements.txt', '--without-hashes']
+            subprocess.run(export_cmd_parts)
+            self.using_req_txt = True
+
+        # What requirements are already listed?
         if self.using_req_txt:
             # Build path to requirements.txt.
             self.req_txt_path = f"{self.project_root}/requirements.txt"
@@ -95,6 +117,13 @@ class Command(BaseCommand):
                 requirements = f.readlines()
                 self.requirements = [r.rstrip() for r in requirements]
 
+        if self.using_pipenv:
+            # Build path to Pipfile.
+            self.pipfile_path = f"{self.project_root}/Pipfile"
+
+            # Get list of requirements.
+            self.requirements = self._get_pipfile_requirements()
+
         # Store any heroku-specific settings already in place.
         #   If any have already been written, we don't want to add them again.
         #   This assumes a section at the end, starting with a check for 
@@ -123,6 +152,13 @@ class Command(BaseCommand):
             #   (Substitution for dev branch install address.)
             self._add_req_txt_pkg('django-simple-deploy')
             # self._add_req_txt_pkg('git+git://github.com/ehmatthes/django-simple-deploy')
+        elif self.using_pipenv:
+            # There's nothing to do here. When you use Pipenv to install something,
+            #   it automatically gets added to Pipfile, so simple-deploy should
+            #   already be here. There's no separate step like "pip freeze".
+            # Modify this later if some people are using a Pipenv approach that
+            #   doesn't automatically add installed packages to Pipfile.
+            pass
 
     def _generate_procfile(self):
         """Create Procfile, if none present."""
@@ -150,6 +186,9 @@ class Command(BaseCommand):
         
         if self.using_req_txt:
             self._add_req_txt_pkg('gunicorn')
+        elif self.using_pipenv:
+            # Here.
+            self._add_pipenv_pkg('gunicorn')
 
 
     def _check_allowed_hosts(self):
@@ -202,10 +241,15 @@ class Command(BaseCommand):
     def _add_db_packages(self):
         """Add packages required for the Heroku db."""
         self.stdout.write("    Adding db-related packages...")
+
         # psycopg2 2.9 causes "database connection isn't set to UTC" issue.
         #   See: https://github.com/ehmatthes/heroku-buildpack-python/issues/31
-        self._add_req_txt_pkg('psycopg2<2.9')
-        self._add_req_txt_pkg('dj-database-url')
+        if self.using_req_txt:
+            self._add_req_txt_pkg('psycopg2<2.9')
+            self._add_req_txt_pkg('dj-database-url')
+        elif self.using_pipenv:
+            self._add_pipenv_pkg('psycopg2', version="<2.9")
+            self._add_pipenv_pkg('dj-database-url')
 
 
     def _add_db_settings(self):
@@ -243,7 +287,10 @@ class Command(BaseCommand):
         # Add whitenoise to requirements.
         self.stdout.write("\n  Configuring static files for Heroku deployment...")
         self.stdout.write("    Adding staticfiles-related packages...")
-        self._add_req_txt_pkg('whitenoise')
+        if self.using_req_txt:
+            self._add_req_txt_pkg('whitenoise')
+        elif self.using_pipenv:
+            self._add_pipenv_pkg('whitenoise')
 
         # Modify settings.
         # DEV: There are three lines here; this can easily be refactored.
@@ -321,6 +368,8 @@ class Command(BaseCommand):
         msg += "\n- Push the changes to Heroku."
         msg += "\n- Migrate the database on Heroku."
         msg += "\n\nThe following commands should finish your initial deployment:"
+        if self.using_pipenv:
+            msg += "\n$ pipenv lock"
         msg += "\n$ git add ."
         msg += '\n$ git commit -am "Configured for Heroku deployment."'
         msg += "\n$ git push heroku main"
@@ -332,6 +381,33 @@ class Command(BaseCommand):
 
     # --- Utility methods ---
 
+    def _get_pipfile_requirements(self):
+        """Get a list of requirements that are already in the Pipfile."""
+        with open(self.pipfile_path) as f:
+            lines = f.readlines()
+
+        requirements = []
+        in_packages = False
+        for line in lines:
+            # Ignore all lines until we've found the start of packages.
+            #   Stop parsing when we hit dev-packages.
+            if '[packages]' in line:
+                in_packages = True
+                continue
+            elif '[dev-packages]' in line:
+                # Ignore dev packages for now.
+                break
+
+            if in_packages:
+                pkg_name = line.split('=')[0].rstrip()
+
+                # Ignore blank lines.
+                if pkg_name:
+                    requirements.append(pkg_name)
+
+        return requirements
+
+
     def _add_req_txt_pkg(self, package_name):
         """Add a package to requirements.txt, if not already present."""
         root_package_name = package_name.split('<')[0]
@@ -344,10 +420,47 @@ class Command(BaseCommand):
             self.stdout.write(f"    Found {root_package_name} in requirements file.")
         else:
             with open(self.req_txt_path, 'a') as f:
-                f.write(f"\n{package_name}    # Added by simple_deploy command.")
+                # DEV: Align these comments like they're aligned in Pipfile.
+                # Align comments, so we don't make req_txt file ugly.
+                #   Version specs are in package_name in req_txt approach.
+                tab_string = ' ' * (30 - len(package_name))
+                f.write(f"\n{package_name}{tab_string}# Added by simple_deploy command.")
 
             self.stdout.write(f"    Added {package_name} to requirements.txt.")
 
+    def _add_pipenv_pkg(self, package_name, version=""):
+        """Add a package to Pipfile, if not already present."""
+        pkg_present = any(package_name in r for r in self.requirements)
+
+        if pkg_present:
+            self.stdout.write(f"    Found {package_name} in Pipfile.")
+        else:
+            self._write_pipfile_pkg(package_name, version)
+
+
+    def _write_pipfile_pkg(self, package_name, version=""):
+        """Write package to Pipfile."""
+        # Write package name right after [packages].
+        #   For simple projects, this shouldn't cause any issues.
+        #   If ordering matters, we can address that later.
+        with open(self.pipfile_path) as f:
+            pipfile_text = f.read()
+
+        if not version:
+            version = '*'
+
+        # Don't make ugly comments; make space to align comments.
+        #   Align at column 30, so take away name length, and version spec space.
+        tab_string = ' ' * (30 - len(package_name) - 5 - len(version))
+
+        new_pkg_string = f'[packages]\n{package_name} = "{version}"{tab_string}# Added by simple_deploy command.'
+        pipfile_text = pipfile_text.replace("[packages]", new_pkg_string)
+
+        with open(self.pipfile_path, 'w') as f:
+            f.write(pipfile_text)
+
+        self.stdout.write(f"    Added {package_name} to Pipfile.")
+
 
     def _check_current_heroku_settings(self, heroku_setting):
         """Check if a setting has already been defined in the heroku-specific
